; For my small son who loves trains. Toot Toot! :-)
; Oh boy, I spent a LOT of time on code optimizations in this prod...
;
; It started as attempt to demake "Immediate Railways" by Digimind and went into very other direction.
; Other inspiration was old BASIC program called "Atari Choo-Choo" from Atari Owner's Manual. As 6 years old I was so disappointed
; that this super-long (at the time) program that I typed was doing nothing more than some simple train sound and printing "toot" on screen :)
;
; How it works?
; - Screen is made of sprites (clouds), beam-sync color changes for gradients, text mode for the train/smoke and the gfx mode for the track.
; - There is very high data reuse (e.g. color values are characters of smoke, audio values are pointers to ROM for pseudo-random gradient etc.) 
; - Vertical scrolling part is synced with track drawing
; - Horizontal scrolling is moving cloud sprites in parallax, cactus sprite and is setting hardware scrolling of track line
; - Train is drawn using OS print function, however Display List is modified to center train instead of having it on the left side of the screen.
; - Toot sound is made of 2 channels for sounding more realistic
; - There is huge reuse of state of registers at different time of code execution

SAFE_VERSION = 0 ; +4 bytes

TRACK_CHAR equ %10010101 ; for gfx mode
TRAIN_LINE_ADDRESS equ SMOKE_LINE_ADDRESS + 3*40 - 7

ANTIC2 equ $02
ANTIC4_HSCROLL equ $14
ANTICF_HSCROLL equ $1F

DL_MAiN_LMS_ADRESS equ $bb84 ; for moving train to center

DL_TRACK_ADRESS equ $bb93 ; dlist address to modify
TRACK_LINE_ADDRESS equ $4000-5
SMOKE_LINE_ADDRESS equ $BD00

PMG_GFX equ $CC14
PMG_GFX_END equ $CC1D
PMG_SIZE equ PMG_GFX_END - PMG_GFX

; these can be probably nicely initialized
AUDF1	equ $D200
AUDC1	equ $D201
AUDF2	equ $D202
AUDC2	equ $D203
AUDF3	equ $D204
AUDC3	equ $D205
AUDF4	equ $D206
AUDC4	equ $D207 
AUDCTL	equ $D208
; 9, A, B, 
; C, D, E
SKCTL   equ $D20F          	

WSYNC	equ $D40a
HSCROL  equ $D404

TIMER1  equ $13 ; slower one
TIMER2  equ $14 ; faster one

ROWCRS	equ $54	;Row of cursor, 1 byte
COLCRS	equ $55	;Column of cursor, 2 bytes

SAVMSC equ $58

OLDROW  equ $5A
OLDCOL  equ $5B ; 2 bytes

ADRESS equ $64
RAMTOP equ $6A

HPOSP0 = $D000 ; Player 0 Horizontal Position
HPOSP1 = $D001 ; Player 1 Horizontal Position
HPOSP2 = $D002 ; Player 2 Horizontal Position
HPOSP3 = $D003 ; Player 3 Horizontal Position
;
HPOSM0 = $D004 ; Missile 0 Horizontal Position
HPOSM1 = $D005 ; Missile 1 Horizontal Position
HPOSM2 = $D006 ; Missile 2 Horizontal Position
HPOSM3 = $D007 ; Missile 3 Horizontal Position
;
SIZEP0 = $D008 ; Player 0 Size
SIZEP1 = $D009 ; Player 1 Size
SIZEP2 = $D00A ; Player 2 Size
SIZEP3 = $D00B ; Player 3 Size
SIZEM =  $D00C ; Missiles Sizes (2 bits per missile)
;
GRAFP0 = $D00D ; Player 0 Graphics Pattern
GRAFP1 = $D00E ; Player 1 Graphics Pattern
GRAFP2 = $D00F ; Player 2 Graphics Pattern
GRAFP3 = $D010 ; Player 3 Graphics Pattern
GRAFM =  $D011 ; Missile Graphics Pattern (2 bits per missile)

COLPF0	equ $D016
COLPF1	equ $D017
COLPF2	equ $D018
COLPF3	equ $D019
COLBK	equ $D01A

VCOUNT	equ $D40B

SDMCTL  equ $022F          	
GPRIOR  equ $026F ; PRIOR - Control Priority, Fifth Player and GTIA modes

PCOLR0	equ $02C0				;p/m 0 color
PCOLR1	equ $02C1				;p/m 1 color
PCOLR2	equ $02C2				;p/m 2 color
PCOLR3	equ $02C3				;p/m 3 color

COLOR0	equ $02c4
COLOR1	equ $02c5
COLOR2	equ $02c6
COLOR3	equ $02c7
COLOR4	equ $02c8

; Aliases, because orig reg names are hard to remember ;)
cursor_y    equ ROWCRS
cursor_x    equ COLCRS
prev_y      equ OLDROW
prev_x      equ OLDCOL

; OS functions
openmode  equ $ef9c
print     equ $f1a4

screen_y equ $0
scroll_value equ $7e
smoke_delta equ $7f ; temp variable
	org $80

.IF SAFE_VERSION = 0
	bvc start
.ENDIF

; train by Piesiu
train:	
	.byte $8A, $10, $60, $60
	.byte $9B
	.byte $C4, $A3, $A3, $93
	.byte $9B
	.byte $20, $93, $85, $97

audio_table:
	; 3D - lower tracks, also $5D works - $5D has LMS and HSCROL
	; 3A - higher tracks
	.byte $5D ; AUDF2 - toot1, last value in loop is used as A register to modify DLIST
	.byte 0  ; AUDC2 
rom_location:
	.byte $40 ; AUDF3 - toot2 ; $40 hex
	.byte $f0 ; AUDC3 not used in initialization so we USE this byte as pointer to ROM ($F040)
smoke_location: ; $BD00
	.byte 0  ; AUDF4 - pitch
	.byte $Bd  ; AUDC4 - disabled volume
	.byte 0 ; AUDCTL - 0 = no mixing of channel
color_table:
	.byte $0e ; PCOLR2
train_index: ; gets zeroed when train is drawn, but it's after initializations
	.byte $0d ; PCOLR3
;piesiu original colors
;	.byte $04 ; COLOR0
;	.byte $77 ; COLOR1
;	.byte $08  ; COLOR2
;	.byte $34  ; COLOR3

;my colors are at the same time chars for the smoke

sky_hi_location: ; $7b04
	.byte $04 ; COLOR0
	.byte $7B ; COLOR1
	.byte $07 ; COLOR2
track_x: ; this is getting zeroed with draw_track, used also for initial scrolling of screen
;.IF SAFE_VERSION=1
;	.byte $34  ; COLOR3
;.ELSE
	.byte $30  ; COLOR3
;.ENDIF

smoke_chars equ track_x-3

; next one must be proper e.g. .byte 1 ; FOR STIMER, otherwise Toot sound is wrong
smoke_index: ; 1,2,$2b,$53
	.byte $07
sky_lo_location: ;$3209
	.byte $08
	.byte $31
	.byte $5A
	;;;;;;;;;;;;;;;
.print "*** Data: ", *-train

start:
	lda #12  ; antic 4, for track gfx	
	jsr openmode	; 3 bytes, Changes $7B (SWPFLG) on zero page to zero so we can't load data there
	; A=UNDEF X=mode Y=01

	; for drawing train
	stx cursor_y
	sty SKCTL ; SKCTL - init sound for proper loading from DOS	
.local init_colors_and_audio
	lda #$FF
	; to initialize P0 i p1 make sta SIZEP2-4,x  
	sta SIZEP2,x ; init SIZEP2, SIZEP3, GRAFP2, GRAFP3 to $FF
	lda audio_table,x
	sta AUDF2,x
	sta PCOLR2-7,x
	sta DL_TRACK_ADRESS,x
	sta TIMER1-4,x
	dex ; 1 byte
rom_location2:
	bpl init_colors_and_audio ; 2 bytes
.endl

	;A = $5D
	;X = $FF
	;Y = $01
				
infi_loop:
	lda VCOUNT	; wait for vcount = 0
	bne infi_loop
	; A=0
	sta HPOSP0	
	
	;;; !OPT CHECK IF SHORTER WHEN OUR VCOUNT REPLACEMENT
.local screen_loop ; place clouds and gradient
	lda VCOUNT
	clc          
	adc track_x 
	
	cmp #118
	beq exit	

	asl	; dense lines
	sta GRAFP0
	tay 
		
	; gradient
	lsr
	lsr
	lsr

	; sky
	adc #$7f
	cpy #$80
	bcc store
	
;;;;;;;; ground

	sta val

	; add bush
	lda #0
	cpy #$8e
	bpl no_pos
	lda sky_pos_hi
no_pos:
	sta HPOSP0	

	tya

	sbc (init_colors_and_audio.rom_location2),y	; add "random" rom value - 1 byte shorter version of "adc $F048,y"	 
;	sbc $F018,y
;	sbc (rom_location),y	; add "random" rom value - 1 byte shorter version of "adc $F048,y"	 
;	sbc (screen_loop.rom_location3-1),y	; add "random" rom value - 1 byte shorter version of "adc $F048,y"	 

	and #%11
	eor #$8F ; set colors
	sbc #0
val equ *-1	
	
	; hide pmg on ground
	; must be at least #122 to be static (up to this we scroll)
	ldy #$B6 ; $F040+$B6 - high value of the byte is to move PMG out of screen to the right

store
	tax ; keep A in X for STX COLBK

	; change position of Cloud	
;	clc ; !!!OPT consider removing 
	lda (sky_hi_location),y
	adc (rom_location),y	; add "random" rom value - 1 byte shorter version of "adc $F048,y"	 
	sta wsync
	stx COLBK
	; changes must be done shortly after wsync
	sta HPOSP2
	adc #31 ; move other player
	sta HPOSP3
	
	jmp screen_loop ; !!!OPT check if somehow can be replaced by branch
exit:						
.endl

;********** post_screen_processing
	;A = $80
	;X = $F1
	;Y = $B6 

;	ldy #$b6 ; center
;	ldy #$c8 ; left


;???	ldy #$a4 ; right
;???train_pos equ *-1
	sty DL_MAiN_LMS_ADRESS ; reposition train to the center of the screen

.local engine
;	reg A is constant so we can just do SBC TIMER2
	sbc TIMER2
	and #7 ; volume, the rest is noise
	sta AUDC4 ; 0-7
	sta scroll_value
.endl

.local draw_track
	bne skip_track_draw ; slower drawing 
	inc smoke_delta
	ldy track_x
	beq skip_track_draw
	lda #TRACK_CHAR ; track
;.IF SAFE_VERSION = 0
;	sta TRACK_LINE_ADDRESS,y
;.ELSE
	sta (rom_location-1),y
;.ENDIF
	dec track_x	
.endl
.print "*** draw_track", .len draw_track

skip_track_draw:

	lda TIMER1
		
	cmp #4
branch_infi1:
	bmi infi_loop

	; A is still TIMER1
	cmp #7 ; longer pause	
branch_infi2:
	bmi scroll

.local toot
	cmp #9
	bmi end_toot
	
	ldy #(7 + $80+$20) ; value to play sound on AUDCx
	cpy TIMER2
	bmi store
	ldy #0
store:
	sty AUDC2
	sty AUDC3
end_toot:
.endl

	
.local draw_train
	; no need to set cursor_x, it's our first OS print now
loop_char:
	ldx train_index
	bmi skip_drawing_train
	lda train,x
	jsr print ; check what regs it's changing
	dec train_index
.endl

skip_drawing_train:
		
;move_train:
;;	bpl end_move_train ; don't move train until drawn
;	ldy train_pos
;	cpy #$b6
;	beq end_move_train
;	inc train_pos	
;end_move_train:

.local add_smoke
	lda TIMER2
	and #3
	tax	
	ldy smoke_index,x
	; get char
	adc smoke_delta
	and #3
	tax
	lda smoke_chars,x
	sta (smoke_location),y
.endl

	; scroll clouds and track
.local scroll
	lda scroll_value
	sta HSCROL

	ldy #127 ; MAX PMG to scroll - #122 is fine
y_loop:
	sty screen_y
	lda (sky_lo_location),y
	adc screen_y
	sta (sky_lo_location),y

	lda (sky_hi_location),y
	sbc #0
	sta (sky_hi_location),y
	dey
	bpl y_loop
.endl	
	bmi branch_infi1
;	jmp infi_loop
		
	org $7b04 ; sky_hi_location
sky_pos_hi: .ds [256]
	org $3209 ; sky_lo_location
sky_pos_lo: .ds [256]

.IF SAFE_VERSION = 1
	RUN start
.ENDIF
			
